/*
 * Aimject
 *
 * Copyright (C) 2006 Jon Oberheide <jon@oberheide.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <dnet.h>
#include <time.h>

#include "main.h"
#include "aim.h"
#include "util.h"
#include "interface.h"

extern aimject_t *aj;

void
put_static(gchar **buf, gpointer tmp, gint bytes)
{
	memcpy(*buf, tmp, bytes);
	*buf += bytes;
}

void
put_uint32(gchar **buf, guint32 tmp)
{
	tmp = g_htonl(tmp);
	put_static(buf, &tmp, sizeof(guint32));
}

void
put_uint16(gchar **buf, guint16 tmp)
{
	tmp = g_htons(tmp);
	put_static(buf, &tmp, sizeof(guint16));
}

void
put_uint8(gchar **buf, guint8 tmp)
{
	put_static(buf, &tmp, sizeof(guint8));
}

void
put_flap(gchar **buf, flap_t *flap)
{
	put_uint8(buf, flap->cmd);
	put_uint8(buf, flap->channel);
	put_uint16(buf, flap->seqnum);
	put_uint16(buf, flap->length);
}

void
put_fnac(gchar **buf, fnac_t *fnac)
{
	put_uint16(buf, fnac->family);
	put_uint16(buf, fnac->subtype);
	put_uint16(buf, fnac->flags);
	put_uint32(buf, fnac->id);
}

void
put_tlv(gchar **buf, tlv_t *tlv)
{
	put_uint16(buf, tlv->type);
	put_uint16(buf, tlv->length);
	put_static(buf, tlv->value, tlv->length);
}

void
get_null(gchar **buf, gint *len, gint bytes)
{
	*buf += bytes;
	*len -= bytes;	
}

void
get_static(gchar **buf, gint *len, gpointer tmp, gint bytes)
{
	memcpy(tmp, *buf, bytes);
	*buf += bytes;
	*len -= bytes;
}

gchar *
get_buf(gchar **buf, gint *len, gint bytes)
{
	gchar *tmp = g_new0(gchar, bytes);
	get_static(buf, len, tmp, bytes);
	return tmp;
}

gchar *
get_string(gchar **buf, gint *len, gint bytes)
{
	gchar *tmp = g_new0(gchar, bytes + 1);
	get_static(buf, len, tmp, bytes);
	return tmp;
}

guint32
get_uint32(gchar **buf, gint *len)
{
	guint32 tmp;
	get_static(buf, len, &tmp, sizeof(tmp));
	tmp = g_ntohl(tmp);
	return tmp;
}

guint16
get_uint16(gchar **buf, gint *len)
{
	guint16 tmp;
	get_static(buf, len, &tmp, sizeof(tmp));
	tmp = g_ntohs(tmp);
	return tmp;
}

guint8
get_uint8(gchar **buf, gint *len)
{
	guint8 tmp;
	get_static(buf, len, &tmp, sizeof(tmp));
	return tmp;
}

flap_t *
get_flap(gchar **buf, gint *len)
{
	flap_t *flap;

	g_return_val_if_fail(*len >= FLAP_LEN, NULL);

	flap = g_new0(flap_t, 1);
	flap->ptr = *buf;
	flap->cmd = get_uint8(buf, len);
	flap->channel = get_uint8(buf, len);
	flap->seqnum = get_uint16(buf, len);
	flap->length = get_uint16(buf, len);

	return flap;
}

void
free_flap(flap_t *flap)
{
	g_free(flap);
}

fnac_t *
get_fnac(gchar **buf, gint *len)
{
	fnac_t *fnac;
	
	g_return_val_if_fail(*len >= FNAC_LEN, NULL);

	fnac = g_new0(fnac_t, 1);
	fnac->ptr = *buf;
	fnac->family = get_uint16(buf, len);
	fnac->subtype = get_uint16(buf, len);
	fnac->flags = get_uint16(buf, len);
	fnac->id = get_uint32(buf, len);

	return fnac;
}

void
free_fnac(fnac_t *fnac)
{
	g_free(fnac);
}

tlv_t *
get_tlv(gchar **buf, gint *len)
{
	tlv_t *tlv;

	g_return_val_if_fail(*len >= sizeof(guint16) + sizeof(guint16), NULL);

	tlv = g_new0(tlv_t, 1);
	tlv->ptr = *buf;
	tlv->type = get_uint16(buf, len);
	tlv->length = get_uint16(buf, len);

	if (tlv->length > *len) {
		g_free(tlv);
		return NULL;
	}

	tlv->value = get_buf(buf, len, tlv->length);

	return tlv;
}

void
free_tlv(tlv_t *tlv)
{
	g_free(tlv->value);
	g_free(tlv);
}

gboolean
inject_aim_msg(client_t *client, conv_t *conv, gchar *text, gint direction)
{
	gchar *packet, *ptr, *font, *sn, *data;
	gint packet_len, extra_len;
	gint fnac_type;
	GIOChannel *io;
	guint16 next_seq;
	guint32 next_fnac; 
	GIOStatus status;
	gsize bytes;

	g_debug("INJECTING |%s| |%s|", conv->sn_norm, text);
	g_debug("last_in: %d  adj_in: %d  last_out: %d  adj_out: %d  last_fnac_out: %d", client->last_seq_in, client->adj_seq_in, client->last_seq_out, client->adj_seq_out, client->last_fnac_out);

	/* account for warning level and tlv count in incoming messages */
	if (direction == DIR_IN) {
		sn = conv->sn_fmt;
		extra_len = 4;
		font = conv->font;
		fnac_type = FNAC_SUB_INCOMING;
		io = client->bos_conn->cio;
		next_seq = client->last_seq_in + 1;
		next_fnac = rand_uint32(aj->rand);
	} else {
		sn = conv->sn_norm;
		extra_len = 0;
		font = client->font;
		fnac_type = FNAC_SUB_OUTGOING;
		io = client->bos_conn->rio;
		next_seq = client->last_seq_out + 1;
		next_fnac = client->last_fnac_out + 1;
	}

	data = g_strdup_printf("%s%s", (font == NULL ? "" : font), text);

	packet_len = FLAP_LEN + \
		     FNAC_LEN + \
		     ICBM_COOKIE_LEN + \
		     MSG_CHANNEL_ID_LEN + \
		     1 + strlen(sn) +			/* screen name LV */ \
		     TLV_TYPE_LEN + TLV_LENGTH_LEN +	/* message block TL */ \
		     TLV_MESSAGE_LEN +			/* message block V */ \
		     strlen(data) + \
		     extra_len;


	packet = g_new0(gchar, packet_len);
	ptr = packet;

	/* flap */
	put_uint8(&ptr, FLAP_CMD_START);
	put_uint8(&ptr, FLAP_FNAC_DATA);
	put_uint16(&ptr, next_seq);
	put_uint16(&ptr, packet_len - FLAP_LEN);
	g_debug("using seqnum: %d", next_seq);

	/* fnac */
	put_uint16(&ptr, FNAC_FAM_MESSAGE);
	put_uint16(&ptr, fnac_type);
	put_uint16(&ptr, 0x0000);
	put_uint32(&ptr, next_fnac);
	g_debug("using fnac id: %d", next_fnac);

	/* icbm */
	rand_get(aj->rand, ptr, ICBM_COOKIE_LEN);
	ptr[ICBM_COOKIE_LEN-1] = '\0';
	ptr += ICBM_COOKIE_LEN;
	put_uint16(&ptr, MSG_CHANNEL_ID);

	/* screenname */
	put_uint8(&ptr, strlen(sn));
	memcpy(ptr, sn, strlen(sn));
	ptr += strlen(sn);

	if (direction == DIR_IN) {
		/* unknown */
		put_uint16(&ptr, 0x0000);

		/* offset to message tlv */
		put_uint16(&ptr, 0x0000);
	}

	/* message block tlv */
	put_uint16(&ptr, TLV_MESSAGE);
	put_uint16(&ptr, TLV_MESSAGE_LEN + strlen(data));

	/* features tlv */
	put_uint16(&ptr, TLV_FEATURES);
	put_uint16(&ptr, TLV_FEATURES_LEN);
	put_uint32(&ptr, TLV_FEATURES_VAL);
	
	/* block type/length */
	put_uint16(&ptr, TLV_MESSAGE_BLOCK);
	put_uint16(&ptr, TLV_TYPE_LEN + TLV_LENGTH_LEN + strlen(data));

	/* block char set/subset */
	put_uint16(&ptr, 0x0000);
	put_uint16(&ptr, 0x0000);

	/* message */
	memcpy(ptr, data, strlen(data));
	g_free(data);

	/* bump our sequence number to keep it synchronized */
	if (direction == DIR_IN) {
		client->adj_seq_in++;
		client->last_seq_in++;
	} else {
		client->adj_seq_out++;
		client->last_seq_out++;
		client->last_fnac_out++;
	}

	if (!(g_io_channel_get_flags(io) & G_IO_FLAG_IS_WRITEABLE)) {
		g_debug("not writeable!");
		return FALSE;
	}

	status = g_io_channel_write_chars(io, packet, packet_len, &bytes, NULL);
	g_debug("wrote %zd bytes to the wire, status %d", bytes, status);
	switch (status) {
		case G_IO_STATUS_AGAIN:
		case G_IO_STATUS_ERROR:
			g_debug("injection: socket error");
			return FALSE;
		break;

		case G_IO_STATUS_EOF:
			g_debug("injection: socket eof");
			return FALSE;
		break;

		case G_IO_STATUS_NORMAL:
			/* all is well */
		break;
	}

	return TRUE;
}

gboolean
dissect_aim_msg(client_t *client, gchar *packet, gint packet_len, gint direction)
{
	flap_t *flap;
	fnac_t *fnac;
	tlv_t *tlv;
	gchar *sn_fmt, *sn_norm;
	guint8 sn_len;
	gboolean muted = FALSE;

	g_debug("dissecting aim packet of length %d", packet_len);

	gchar *next_flap = packet;
	gint flap_bytes_left;
	gint packet_bytes_left = packet_len;

	while (TRUE) {
		conv_t *conv;
		msg_t *msg = NULL;
		gchar *msg_data = NULL;
		gboolean auto_resp = FALSE;

		/* get flap */
		flap = get_flap(&next_flap, &packet_bytes_left);
		goto_if_fail(flap != NULL, out);

		g_debug("flap cmd %u", flap->cmd);
		g_debug("flap channel %u", flap->channel);
		g_debug("flap seqnum %u", flap->seqnum);
		g_debug("flap length %u", flap->length);

		/* adjust packet ptrs and lengths */
		packet = next_flap;
		next_flap += flap->length;
		flap_bytes_left = flap->length;
		packet_bytes_left -= flap->length;

		/* if its not aim, ignore */
		goto_if_fail(flap->cmd == FLAP_CMD_START, flap_cleanup);

		/* synchronize flap sequence numbers */
		guint16 seq_adjust;
		if (direction == DIR_IN) {
			seq_adjust = client->adj_seq_in;
			client->last_seq_in = flap->seqnum + client->adj_seq_in;
			g_debug("last seq in set to %u and adj in is %u", client->last_seq_in, client->adj_seq_in);
		} else {
			seq_adjust = client->adj_seq_out;
			client->last_seq_out = flap->seqnum + client->adj_seq_out;
			g_debug("last seq out set to %u and adj out is %u", client->last_seq_out, client->adj_seq_out);
		}
		flap->seqnum += seq_adjust;
		put_flap(&flap->ptr, flap);
		g_debug("seqnum changed to %u", flap->seqnum);

		/* if its not fnac, ignore */
		goto_if_fail(flap->channel == FLAP_FNAC_DATA, flap_cleanup);

		/* if theres not enough packet left for a fnac, ignore */
		goto_if_fail(packet_bytes_left + flap->length >= FNAC_LEN, flap_cleanup);

		/* get fnac */
		fnac = get_fnac(&packet, &flap_bytes_left);
		goto_if_fail(fnac != NULL, flap_cleanup);

		g_debug("fnac family %u", fnac->family);
		g_debug("fnac subtype %u", fnac->subtype);
		g_debug("fnac flags %u", fnac->flags);
		g_debug("fnac id %u", fnac->id);

		/* synchronize fnac ids */
		if (direction == DIR_OUT) {
			client->last_fnac_out = fnac->id + client->adj_seq_out;
			g_debug("last fnac out set to %u and adj out is %u", client->last_fnac_out, client->adj_seq_out);
			fnac->id += client->adj_seq_out;
			put_fnac(&fnac->ptr, fnac);
		}

		g_debug("packleft: %d, flaplen: %d, fnaclen: 10, flapleft: %d", packet_bytes_left, flap->length, flap_bytes_left);
		goto_if_fail(packet_bytes_left + flap->length - FNAC_LEN >= flap_bytes_left, fnac_cleanup);
		goto_if_fail(fnac->family == FNAC_FAM_MESSAGE, fnac_cleanup);
		goto_if_fail(fnac->subtype == FNAC_SUB_INCOMING || fnac->subtype == FNAC_SUB_OUTGOING, fnac_cleanup);

		/* skip icbm cookie (8 bytes) */
		get_null(&packet, &flap_bytes_left, ICBM_COOKIE_LEN);

		/* skip msg channel id (2 bytes) */
		get_null(&packet, &flap_bytes_left, MSG_CHANNEL_ID_LEN);

		/* get buddy screenname */
		sn_len = get_uint8(&packet, &flap_bytes_left);
		goto_if_fail(sn_len > 0 && sn_len <= MAX_SN_LEN, fnac_cleanup);
		sn_fmt = get_string(&packet, &flap_bytes_left, sn_len);

		/* normalize screenname */
		sn_norm = str_normalize(sn_fmt);

		/* incoming msgs have a bit extra data */
		if (fnac->subtype == FNAC_SUB_INCOMING) {
			get_null(&packet, &flap_bytes_left, 2);
			get_null(&packet, &flap_bytes_left, 2);
		}

		/* get tlvs */
		while ((tlv = get_tlv(&packet, &flap_bytes_left)) != NULL) {
			g_debug("got a tlv of type %d length %d", tlv->type, tlv->length);

			if (tlv->type == TLV_AUTO_RESP) {
				g_debug("auto response tlv found");
				auto_resp = TRUE;
			}

			if (tlv->type == TLV_MESSAGE) {
				tlv_t *subtlv;
				gint tlv_len = tlv->length;
				gchar *tlv_val = tlv->value;

				g_debug("message tlv found");

				/* inside a msg TLV are sub-TLVs */
				while ((subtlv = get_tlv(&tlv_val, &tlv_len)) != NULL) {
					g_debug("got a subtlv of type %d length %d", subtlv->type, subtlv->length);

					if (subtlv->type == TLV_MESSAGE_BLOCK) {
						gint subtlv_len = subtlv->length;
						gchar *subtlv_val = subtlv->value;

						get_null(&subtlv_val, &subtlv_len, 2);
						get_null(&subtlv_val, &subtlv_len, 2);
						msg_data = get_string(&subtlv_val, &subtlv_len, subtlv_len);

						g_debug("message block tlv found");
						g_debug("sn: |%s| msg: |%s|", sn_norm, msg_data);
					}

					free_tlv(subtlv);
				}
			}

			free_tlv(tlv);
		}

		/* determine if a message was present in the flap */
		if (msg_data) {
			msg = g_new0(msg_t, 1);
			msg->type = (auto_resp ? MSG_AUTO_RESP : MSG_NONE);
			msg->dir = (fnac->subtype == FNAC_SUB_INCOMING ? DIR_IN : DIR_OUT);
			msg->timestamp = time(NULL);
			msg->data = str_strip_html(msg_data);

			/* add message to appropriate conv */
			conv = main_lookup_conv(client, sn_norm);
			if (!conv) {
				conv = main_add_conv(client, sn_norm, sn_fmt);
				interface_add_conv(client, conv);
			} else {

			}

			/* muting support */
			if (direction == DIR_IN && conv->mute_from) {
				muted = TRUE;
				msg->type = MSG_MUTED;
				client->adj_seq_in--;
				client->last_seq_in--;
			} else if (msg->dir == DIR_OUT && conv->mute_to) {
				muted = TRUE;
				msg->type = MSG_MUTED;
				client->adj_seq_out--;
				client->last_seq_out--;
				client->last_fnac_out--;
			}

			/* update html font styles */
			if (msg->dir == DIR_IN) {
				g_free(conv->font);
				conv->font = str_get_font(msg_data);

				/* also update sn formatting */
				if (strcmp(sn_fmt, conv->sn_fmt) != 0) {
					g_free(conv->sn_fmt);
					conv->sn_fmt = g_strdup(sn_fmt);
					g_debug("conv sn formatting set to |%s|", conv->sn_fmt);
				}
			} else {
				g_free(client->font);
				client->font = str_get_font(msg_data);
			}

			main_add_msg(client, conv, msg);
			interface_add_msg(client, conv, msg);

			g_free(msg_data);
		}

		g_free(sn_norm);
		g_free(sn_fmt);

		fnac_cleanup:
			free_fnac(fnac);
		flap_cleanup:
			free_flap(flap);
	}

	out:
		return muted;
}

gint
dissect_aim_bos(client_t *client, gchar *packet, gint packet_len)
{
	flap_t *flap;
	fnac_t *fnac;
	tlv_t *tlv;

	g_debug("dissecting aim packet of length %d", packet_len);

	gchar *tmp;
	gchar *next_flap = packet;
	gint flap_bytes_left;
	gint packet_bytes_left = packet_len;

	while (TRUE) {
		/* get flap */
		flap = get_flap(&next_flap, &packet_bytes_left);
		goto_if_fail(flap != NULL, out);

		g_debug("flap cmd %u", flap->cmd);
		g_debug("flap channel %u", flap->channel);
		g_debug("flap seqnum %u", flap->seqnum);
		g_debug("flap length %u", flap->length);

		/* adjust packet ptrs and lengths */
		packet = next_flap;
		next_flap += flap->length;
		flap_bytes_left = flap->length;
		packet_bytes_left -= flap->length;

		/* if its not aim, ignore */
		goto_if_fail(flap->cmd == FLAP_CMD_START, flap_cleanup);

		/* if its not fnac, ignore */
		goto_if_fail(flap->channel == FLAP_FNAC_DATA, flap_cleanup);

		/* if theres not enough packet left for a fnac, ignore */
		goto_if_fail(packet_bytes_left + flap->length >= FNAC_LEN, flap_cleanup);

		/* get fnac */
		fnac = get_fnac(&packet, &flap_bytes_left);
		goto_if_fail(fnac != NULL, flap_cleanup);

		g_debug("fnac family %u", fnac->family);
		g_debug("fnac subtype %u", fnac->subtype);
		g_debug("fnac flags %u", fnac->flags);
		g_debug("fnac id %u", fnac->id);

		goto_if_fail(packet_bytes_left + flap->length - FNAC_LEN >= flap_bytes_left, fnac_cleanup);
		goto_if_fail(fnac->family == FNAC_FAM_SIGNON, fnac_cleanup);
		goto_if_fail(fnac->subtype == FNAC_SUB_LOGONREPLY, fnac_cleanup);

		/* get tlvs */
		while ((tlv = get_tlv(&packet, &flap_bytes_left)) != NULL) {
			g_debug("got a tlv of type %d length %d", tlv->type, tlv->length);

			if (tlv->type == TLV_SCREENNAME) {
				tmp = g_strndup(tlv->value, tlv->length);
				client->sn_norm = str_normalize(tmp);
				g_free(tmp);
			}

			if (tlv->type == TLV_BOS) {
				g_debug("tlv bos found, modifying packet");
				g_debug("before tlv - type: %d, length: %d", tlv->type, tlv->length);
				g_debug("before flap - length: %d", flap->length);
				g_debug("before packet - length: %d", packet_len);

				/* save the bos address */
				g_free(client->rhost);
				client->rhost = g_strndup(tlv->value, tlv->length - strlen(":5190"));
				client->state = BOS_CONN;

				/* write changes back to packet */
				g_free(tlv->value);
				tlv->value = g_strdup_printf("%s:5190", ip_ntoa(&aj->intf_pa.addr_ip));
				gint length_diff = strlen(tlv->value) - tlv->length;
				g_debug("length diff: %d", length_diff);

				/* change packet/flap/tlv lengths */
				packet_len += length_diff;
				flap->length += length_diff;
				tlv->length += length_diff;

				/* update packet with new address */
				memmove(packet + length_diff, packet, flap_bytes_left + packet_bytes_left);
				put_flap(&flap->ptr, flap);
				put_tlv(&tlv->ptr, tlv);

				g_debug("after tlv - type: %d, length: %d", tlv->type, tlv->length);
				g_debug("after flap - length: %d", flap->length);
				g_debug("after packet - length: %d", packet_len);
			}

			free_tlv(tlv);
		}

		fnac_cleanup:
			free_fnac(fnac);
		flap_cleanup:
			free_flap(flap);
	}

	out:
		return packet_len;
}
