/*
 * Aimject
 *
 * Copyright (C) 2006 Jon Oberheide <jon@oberheide.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <dnet.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#include "os.h"
#include "conn.h"
#include "interface.h"
#include "aim.h"
#include "main.h"

extern aimject_t *aj;

gboolean conn_io_event(GIOChannel *io, GIOCondition cond, gpointer data);

conn_t *
conn_lookup(client_t *client, GIOChannel *io)
{
	conn_t *conn;

	g_return_val_if_fail(io != NULL, NULL);
	g_return_val_if_fail(client != NULL, NULL);
	g_return_val_if_fail(client->new_conn != NULL, NULL);
	g_return_val_if_fail(client->bos_conn != NULL, NULL);

	conn = client->new_conn;
	if (conn->cio && io == conn->cio) {
		g_debug("found existing conn: %p", conn);
		return conn;
	}
	if (conn->rio && io == conn->rio) {
		g_debug("found existing conn: %p", conn);
		return conn;
	}

	conn = client->bos_conn;
	if (conn->cio && io == conn->cio) {
		g_debug("found existing conn: %p", conn);
		return conn;
	}
	if (conn->rio && io == conn->rio) {
		g_debug("found existing conn: %p", conn);
		return conn;
	}

	return NULL;
}

void
conn_add(client_t *client, gint csock, gint rsock, struct sockaddr_in *addr)
{
	conn_t *conn;
	
	switch (client->state) {
		case NEW_CONN:
			conn = client->new_conn;
			conn->type = NEW_CONN;
		break;

		case BOS_CONN:
			conn = client->bos_conn;
			conn->type = BOS_CONN;
			client->state = DONE;

			/* add client to interface */
			interface_add_client(client);
		break;

		default:
			conn = NULL;
			g_error("unknown state");
		break;
	}

	conn->cio = g_io_channel_new(csock);
	g_io_channel_set_encoding(conn->cio, NULL, NULL);
	g_io_channel_set_buffered(conn->cio, FALSE);
	g_io_channel_set_close_on_unref(conn->cio, TRUE);
	g_io_channel_set_flags(conn->cio, G_IO_FLAG_NONBLOCK, NULL);
	conn->cio_source = g_io_add_watch(conn->cio, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL, conn_io_event, client);

	conn->rio = g_io_channel_new(rsock);
	g_io_channel_set_encoding(conn->rio, NULL, NULL);
	g_io_channel_set_buffered(conn->rio, FALSE);
	g_io_channel_set_close_on_unref(conn->rio, TRUE);
	g_io_channel_set_flags(conn->rio, G_IO_FLAG_NONBLOCK, NULL);
	conn->rio_source = g_io_add_watch(conn->rio, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL, conn_io_event, client);

	memcpy(&conn->addr, addr, sizeof(struct sockaddr_in));

	g_debug("registered new connection: %p", conn);
}

void
conn_destroy(client_t *client, conn_t *conn)
{
	g_source_remove(conn->cio_source);
	g_source_remove(conn->rio_source);

	g_io_channel_shutdown(conn->cio, FALSE, NULL);
	g_io_channel_shutdown(conn->rio, FALSE, NULL);

	g_io_channel_unref(conn->cio);
	g_io_channel_unref(conn->rio);

	conn->cio = NULL;
	conn->rio = NULL;

	g_debug("channels shutdown and conn removed");
}

gboolean
conn_io_hup(client_t *client, conn_t *conn)
{
	conn_destroy(client, conn);

	switch (conn->type) {
		case NEW_CONN:
			g_debug("hang up on new connection detected");
		break;

		case BOS_CONN:
			g_debug("hang up on bos connection detected");

			/* cleanup client */
			interface_destroy_client(client);
			main_destroy_client(client);
		break;

		default:
			g_error("unknown connection type");
		break;
	}

	return FALSE;
}

gboolean
conn_io_in(client_t *client, conn_t *conn, GIOChannel *io)
{
	gsize bytes;
	gssize len;
	gchar buf[8192];
	GIOChannel *io_in, *io_out;
	GIOStatus status;
	gint direction;
	gboolean muted;

	if (io == conn->cio) {
		io_in = conn->cio;
		io_out = conn->rio;
		direction = DIR_OUT;
		g_debug("receiving data from client");
	} else {
		io_in = conn->rio;
		io_out = conn->cio;
		direction = DIR_IN;
		g_debug("receiving data from remote");
	}

	if (!(g_io_channel_get_flags(io_in) & G_IO_FLAG_IS_READABLE)) {
		g_debug("not readable!");
		return FALSE;
	}

	status = g_io_channel_read_chars(io_in, buf, sizeof(buf), &bytes, NULL);
	g_debug("read %zd bytes from the wire, status %d", bytes, status);
	switch (status) {
		case G_IO_STATUS_AGAIN:
		case G_IO_STATUS_ERROR:
			return TRUE;
		break;

		case G_IO_STATUS_EOF:
			return FALSE;
		break;

		case G_IO_STATUS_NORMAL:
			/* all is well */
		break;
	}

	switch (conn->type) {
		case NEW_CONN:
			len = dissect_aim_bos(client, buf, bytes);
		break;

		case BOS_CONN:
			len = bytes;
			muted = dissect_aim_msg(client, buf, bytes, direction);
			
			/* don't relay packet if it's been muted */
			if (muted) {
				g_debug("message muted!");
				return TRUE;
			}
		break;

		default:
			len = bytes;
			g_error("unknown type");
		break;
	}

	if (!(g_io_channel_get_flags(io_out) & G_IO_FLAG_IS_WRITEABLE)) {
		g_debug("not writeable!");
		return FALSE;
	}

	status = g_io_channel_write_chars(io_out, buf, len, &bytes, NULL);
	g_debug("wrote %zd bytes to the wire, status %d", bytes, status);
	switch (status) {
		case G_IO_STATUS_AGAIN:
		case G_IO_STATUS_ERROR:
			return TRUE;
		break;

		case G_IO_STATUS_EOF:
			return FALSE;
		break;
	
		case G_IO_STATUS_NORMAL:
			/* all is well */
		break;
	}

	return TRUE;
}

gboolean
conn_io_event(GIOChannel *io, GIOCondition cond, gpointer data)
{
	client_t *client = (client_t *) data;
	g_return_val_if_fail(client != NULL, FALSE);

	conn_t *conn = conn_lookup(client, io);
	g_return_val_if_fail(conn != NULL, FALSE);

	if (cond & G_IO_IN) {
		g_debug("io_in called");
		return conn_io_in(client, conn, io);
	} else if (cond & (G_IO_ERR | G_IO_HUP | G_IO_NVAL)) {
		g_debug("io_hup called");
		return conn_io_hup(client, conn);
	} else {
		g_warning("unknown gio condition");
	}

	return FALSE;
}

gboolean
conn_sio_in(GIOChannel *sio, GIOCondition cond, gpointer data)
{
	gchar *rhost;
	client_t *client;
	gint ret, csock, rsock, ssock;
	struct sockaddr_in caddr, raddr;
	struct hostent *hostinfo;
	guint ssize = sizeof(struct sockaddr_in);

	ssock = g_io_channel_unix_get_fd(sio);

	csock = accept(ssock, (struct sockaddr *) &caddr, &ssize);
	g_return_val_if_fail(csock != -1, FALSE);

	g_debug("NEW CONNECTION: client %s:%d", inet_ntoa(caddr.sin_addr), g_ntohs(caddr.sin_port));

	/* if client doesnt exists, add it */
	client = main_lookup_client(&caddr);
	if (!client) {
		client = main_add_client();
	}

	/* determine our remote host based on client's state */
	switch (client->state) {
		case NEW_CONN:
			rhost = OSCAR_HOST;
			g_debug("state is new connection");
		break;
	
		case BOS_CONN:
			rhost = client->rhost;
			g_debug("state is bos connection");
		break;

		default:
			rhost = NULL;
			g_error("unknown state");
		break;
	}

	/* look up remote server */
	hostinfo = gethostbyname(rhost);
	g_return_val_if_fail(hostinfo != NULL, FALSE);

	memset(&raddr, 0, ssize);
	raddr.sin_family = AF_INET;
	raddr.sin_addr = *(struct in_addr *) hostinfo->h_addr;
	raddr.sin_port = g_htons(LISTEN_PORT);

	rsock = socket(AF_INET, SOCK_STREAM, 0);
	g_return_val_if_fail(rsock != -1, FALSE);

	g_debug("connect to remote host: %s", rhost);
	ret = connect(rsock, (struct sockaddr *) &raddr, sizeof(raddr));
	g_return_val_if_fail(ret != -1, FALSE);

	conn_add(client, csock, rsock, &caddr);

	return TRUE;
}

void
conn_init(void)
{
	gint ret, ssock;
	struct sockaddr_in saddr;
	gint sockopt_on = 1;
	GIOChannel *sio;
	guint ssize = sizeof(struct sockaddr_in);

	/* set up server socket */
	memset(&saddr, 0, ssize);
	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr = INADDR_ANY;
	saddr.sin_port = g_htons(LISTEN_PORT);

	ssock = socket(AF_INET, SOCK_STREAM, 0);
	g_assert(ssock != -1);

	ret = setsockopt(ssock, SOL_SOCKET, SO_REUSEADDR, (const gpointer) &sockopt_on, sizeof(gint));
	g_assert(ret != -1);

	ret = bind(ssock, (struct sockaddr *) &saddr, sizeof(saddr));
	g_assert(ret != -1);

	ret = listen(ssock, BACKLOG);
	g_assert(ret != -1);

	g_debug("server listening on %s:%d", inet_ntoa(saddr.sin_addr), g_ntohs(saddr.sin_port));

	sio = g_io_channel_new(ssock);
	g_io_channel_set_encoding(sio, NULL, NULL);
	g_io_channel_set_close_on_unref(sio, TRUE);

	/* add watch for incoming connections */
	g_io_add_watch(sio, G_IO_IN, conn_sio_in, NULL);
}
