/*
 * Aimject
 *
 * Copyright (C) 2006 Jon Oberheide <jon@oberheide.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <dnet.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#include "os.h"
#include "main.h"
#include "aim.h"
#include "conn.h"
#include "interface.h"

aimject_t *aj;

void
main_add_msg(client_t *client, conv_t *conv, msg_t *msg)
{
	g_assert(client != NULL);
	g_assert(conv != NULL);
	g_assert(msg != NULL);

	if (msg->dir == DIR_OUT) {
		/* update client statistics */
		client->msg_sent_count++;
		client->last_sent_msg = msg->timestamp;

		g_debug("%s -> %s:  %s", client->sn_norm, conv->sn_norm, msg->data);
	} else {
		/* update client statistics */
		client->msg_recv_count++;
		client->last_recv_msg = msg->timestamp;

		g_debug("%s -> %s:  %s", conv->sn_norm, client->sn_norm, msg->data);
	}

	g_ptr_array_add(conv->msgs, msg);
}

void
main_destroy_msg(conv_t *conv, msg_t *msg)
{
	g_assert(conv != NULL);
	g_assert(msg != NULL);

	g_ptr_array_remove(conv->msgs, msg);
	g_free(msg->data);
	g_free(msg);
}

conv_t *
main_add_conv(client_t *client, gchar *sn_norm, gchar *sn_fmt)
{
	conv_t *conv;

	g_assert(client != NULL);
	g_assert(sn_norm != NULL);
	g_assert(sn_fmt != NULL);

	g_debug("creating new conv from %s to %s", client->sn_norm, sn_norm);

	/* create a new conversion */
	conv = g_new0(conv_t, 1);
	conv->msgs = g_ptr_array_new();
	conv->sn_norm = g_strdup(sn_norm);
	conv->sn_fmt = g_strdup(sn_fmt);

	/* update client statistics */
	client->conv_count++;

	g_ptr_array_add(client->convs, conv);

	return conv;
}

conv_t *
main_lookup_conv(client_t *client, gchar *sn_norm)
{
	guint i;
	conv_t *conv;

	g_assert(client != NULL);
	g_assert(sn_norm != NULL);

	for (i = 0; i < client->convs->len; ++i) {
		conv = g_ptr_array_index(client->convs, i);
		g_assert(conv != NULL);

		if (g_ascii_strcasecmp(sn_norm, conv->sn_norm) == 0) {
			return conv;
		}
	}

	return NULL;
}

void
main_destroy_conv(client_t *client, conv_t *conv)
{
	msg_t *msg;

	g_assert(client != NULL);
	g_assert(conv != NULL);

	g_debug("destroying conv from %s to %s", client->sn_norm, conv->sn_norm);

	/* clean up all messages */
	while (conv->msgs->len > 0) {
		msg = g_ptr_array_index(conv->msgs, 0);
		main_destroy_msg(conv, msg);
	}

	client->conv_count--;

	g_ptr_array_remove(client->convs, conv);
	g_ptr_array_free(conv->msgs, TRUE);
	g_free(conv->sn_norm);
	g_free(conv->sn_fmt);
	g_free(conv->font);
	g_free(conv);
}

client_t *
main_add_client()
{
	client_t *client;
	client = g_new0(client_t, 1);
	client->new_conn = g_new0(conn_t, 1);
	client->bos_conn = g_new0(conn_t, 1);
	client->convs = g_ptr_array_new();
	client->state = NEW_CONN;

	g_debug("registered new client: %p", client);

	g_ptr_array_add(aj->clients, client);

	return client;
}

client_t *
main_lookup_client(struct sockaddr_in *addr)
{
	guint i;
	client_t *client;
	for (i = 0; i < aj->clients->len; ++i) {
		client = g_ptr_array_index(aj->clients, i);
		g_return_val_if_fail(client != NULL, NULL);

		if ((addr->sin_addr.s_addr == client->new_conn->addr.sin_addr.s_addr ||
		     addr->sin_addr.s_addr == client->bos_conn->addr.sin_addr.s_addr) &&
		     client->state != DONE) {
			g_debug("found existing client");
			return client;
		}
	}
	return NULL;
}

void
main_destroy_client(client_t *client)
{
	conv_t *conv;

	g_assert(client != NULL);

	/* clean up all conversations */
	while (client->convs->len > 0) {
		conv = g_ptr_array_index(client->convs, 0);
		main_destroy_conv(client, conv);
	}

	g_ptr_array_remove(aj->clients, client);
	g_ptr_array_free(client->convs, TRUE);
	g_free(client->rhost);
	g_free(client->new_conn);
	g_free(client->bos_conn);
	g_free(client->sn_norm);
	g_free(client->font);
	g_free(client);

	g_debug("client destroyed");
}


void signal_cleanup(gint sig)
{
	g_debug("caught exit signal");
	
	if (aj->arp_loop) {
		g_debug("killing arp main loop...");
		g_main_loop_quit(aj->arp_loop);
		g_debug("arp main loop killed");
	}

	if (aj->arp_thread) {
		g_debug("waiting for arp thread...");
		g_thread_join(aj->arp_thread);
		g_debug("arp thread joined");
	}

	g_debug("terminating aimject...");

	exit(0);
}

gint
main(gint argc, gchar **argv)
{
	/* OS-specific setup */
	os_init();

	aj = g_new0(aimject_t, 1);
	aj->clients = g_ptr_array_new();
	aj->rand = rand_open();

	if (!g_thread_supported()) {
		g_thread_init(NULL);
	}

	/* initialize the interface */
	interface_init(argc, argv);

	gtk_main();

	signal_cleanup(0);

	rand_close(aj->rand);
	g_ptr_array_free(aj->clients, TRUE);
	g_free(aj);

	return 0;
}
